<?php
/**
* XML-LINE Version  0.4.4.4  24.06.2007

* PHP-Klasse zum Abfragen und ndern von XML-Dateien
* (C) 2004-2007 Peter Bieling, pb@media-palette.de
* Einige Codezeilen, fr den Aufruf der Parserfunktionen sind dem
* PHP-Manual entnommen.
******************************************************************
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*************************************************************************
* If you want to write XML and save the XML-Declaration, look if your
* PHP-Version works propery:
* http://bugs.php.net/bug.php?id=27908
*
* @author Peter Bieling <pb@media-palette.de>
*/
class xml_line {
    /**
    * Codierung des Eingabe-XML.
    *
    * @var string $encoding optional
    */
    var $targetencoding="iso-8859-1";

    /**
    * Codierung des Eingabe-XML.
    *
    * @var string $encoding optional
    */
    var $encoding="iso-8859-1";

    /**
    * Der Name der XML-Input-Datei
    *
    * @var string $file
    */
    var $file;

    /**
    * Der Name der Output-Datei
    *
    * @var string $outfile
    */
    var $outfile = "";

    /**
    * Name des aktuellen Elements
    *
    * @var string $akt_name
    */
    var $akt_name = "";

    /**
    * Ausgabe wird nur geschrieben, wenn true gesetzt ist.
    *
    * @var boolean $write_flag
    */
    var $write_flag = true;

    /**
    * Markiert das Ende von Bereichen, die nicht geschrieben (also de facto gelscht) werden.
    *
    * @var string $last_nowrite_line    join("",$this->ls);
    */
    var $last_nowrite_line = "";

    /**
    * Speichert die XML-Ausgabe, sofern sie gefordert ist, bzw. nicht in eine Datei geschrieben wird
    *
    * @var string $xml_buffer;
    */
    var $xml_buffer = "";

    /**
    * Speichert die CSV-Ausgabe, sofern sie (statt XML)gefordert ist,
    * bzw. nicht in eine Datei geschrieben wird. Experimentell!!!
    *
    * @var string $csv_buffer;
    */
    var $csv_buffer = "";

    /**
    * Speichert Position als key und einen String als value fr das Ersetzen
    * oder Einsetzen von Text oder XML .Jede Funktion nutzt diesen Speicher selbst bei Bedarf.
    *
    * @var array $fill     join("", $this->ls)=> Insertstring
    */
    var $fill = array();

    /**
    * Speichert einzusetzende Attribute in einem Hash (name=>value), wenn ein oder mehrere
    * Attribute hinzugefgt werden sollen
    *
    * @var array $fillattr
    */
    var $fillattr = array();

    /**
    * Speichert die Attributhashs solange das Element aktuell ist. (Key ist der "pathname")
    *     Achtung: Wird wegen neuer Funktion get_record zur Zeit nicht gelscht. (todo)
    *     Wichtig fr Einfgeoperationen, damit auch beim Endelement noch bekannt ist,
    *     welche Attribute das Element hat.
    *
    * @var array $temp_attr_buffer
    */
    var $temp_attr_buffer = array();

    /**
    * Gibt an, ob bzw. was ausgegeben werden soll.
    * xml, csv, hixml oder Leerstring
    *
    * @var mixed $output_action
    */
    var $output_action = "";

    /**
    * Gibt den aktuellen Level der Elementhierarchie an
    *
    * @var integer $level
    */
    var $level = 0;

    /**
    * Expat liefert die Daten hufig in mehreren Stcken
    * Hier wird das Ergebnis zwischengespeichert, bis der Data-String komplett ist
    *
    * @var string $data_buffer
    */
    var $data_buffer = "";

    /**
    * Enthlt alle Elemente des aktuellen Elementpfads einzeln als Array-Werte.
    * Damit kann dieser problemlos ergnzt oder gekrzt werden
    *
    * @var array $path_array  (root,level1,level2)
    */
    var $path_array = array();

    /**
    * Name der aufzurufenden Funktion. Funktionen werden in einem Array gespeichert.
    * $action_func[0] = Funktionsname
    * $action_func[1] = Parameterliste (array)
    * $action_func[2] = Kommando
    * $action_func[3] = Ersetzungsstring
    *
    * @var array $action_func
    */
    var $action_func = array();

    /**
    * Zhlt die gesetzten Funktionen. Ermglicht das Auseinanderhalten der verschiedenen Prozesse
    * und die Zuordnung der Rckgabewerte.
    *
    * @var integer $func_counter
    */
    var $func_counter = 0;

    /**
    * Speicher fr zustzliche Prozessinformationen,
    * 'datapart' z.B. ([1]['datapart']) merkt sich bei
    * gemischten Inhalten, welcher Datenabschnitt ausgegeben werden soll
    * 'limit' [1]['limit']=3 bedeutet: 3 Treffer werden gesucht, danach steigt die Funktion aus.
    * (runterzhlen bis auf 0)
    *
    * @var array $procdata
    */
    var $procdata = array(array());

    /**
    * Enthlt die wichtigsten Informationen der aktuellen XML-Position.
    *
    * @var array $ls
    */
    var $ls = array();

    /**
    * Speichert die gefunden Daten, Pfadinformationen und Attribute in einem Array
    * Art der Rckgabe kann sich noch ndern, am besten mit print_r() selbst ansehen. ;-)
    *
    * @var array $found_data
    */
    var $found_data = array();

    /**
    * Zhler fr Textstcke innerhalb eines Elements mit weiteren untergeordneten Elementen,
    * luft nur mit, wenn er bentigt wird und das passende Element gefunden wurde.
    *
    * @var arry $found_data_part (key aus Prozessnr-pathstring-und-levelstring)
    */
    var $found_data_part = array();

    /**
    * Element-Counter. Luft nur mit, wenn er gesetzt wird. Kann fr statistische Zwecke benutzt werden
    *     $el_counter[elname] = array($countall, $level => $countonthislevel
    *
    * @var mixed $elcounter  (string or array)
    */
    var $el_counter = "";

    /**
    * element-level-counter, enthlt die Zhlerwerte, der aktuellen Elemente. [level]["elementname"]
    *
    * @var array $elc
    */
    var $elc = array(array());

    /**
    * Kurzform des aktuellen Elementpfads (3,2,8) ohne Bezug auf die Elemente
    *
    * @var array $elc_path_array
    */
    var $elc_path_array = array();

    /**
    * Error - noch strflich vernachlssigt ;-( Soll die Fehlermeldungen aufnehmen,
    * die man dann abrufen kann. (noch in der Entwicklung)
    *
    * @var array $error
    */
    var $error = array();

    /**
    * Expat-Variable
    *
    * @var integer $parser resource
    */
    var $parser;

    /**
    * Outputfilehandler
    *
    * @var integer $of resource_id
    */
    var $of;

    /**
    * Soll bei Datensatzabfrage, das Ergebnis speichern:
    * $table_result[Funktionsnr][counter]['key'] => "value"
    * key ist die "Spaltenberschrift" = Name des Elements.
    * Ist das Element untergeordnet ist key 'parent/child'
    * Attribute werden ebenfalls in die Tabelle aufgenommen:
    * 'element@attribut1'
    */
    var $table_result = array();

    /**
    * XML-Header und Kommentare ausgeben oder nicht. Wird ber Methodenaufruf verndert
    *
    * @var boolean $ignore_default_data
    */
    var $ignore_default_data = false;
    //var $ignore_default_data = true;
    /**
    * Interner Schalter fr die Highlight-Ausgabe
    *
    * @var boolean $hightlight
    */
    var $highlight = false;

    /**
    * Interner Schalter fr die Highlight-Ausgabe. Merkt sich, ob Daten ausgegeben wurden
    *
    * @var boolean $datafound
    */
    var $datafound = false;

    /**
    * Internes Flag fr die Highlight-Ausgabe. Merkt sich, ob vorher ein Zeilenumbruch gesetzt wurde.
    *
    * @var boolean $breakset
    */
    var $breakset = false;
    // neu:
    var $copy_flag = array(); //zeigt an, ob in  proc-data :Puffer geschrieben werden muss:
    // $copy_flag[0]=true;
    var $tmp_sign = "_%%_"; //temporrer Inhalt fr sonst leere Elemente.
    // wird fr get_record() bentigt.
    var $actioncounter = true;

    var $outer_attr_flag = 0;
    // ==================== TEST ==============
    var $pointer_A = false;
    var $pointer_D = false;
    var $pointer_S = false;
    var $pointer_E = false;
    var $pointer_EOS = false;
    var $pointer_BE = false;
    var $restore_xml_entities = true;
    // =============================================
    /*
     * Wird auf true gesetzt, wenn der Parser einen CDATA-Abschnitt erreicht
     * und auf false, wenn er ihn wieder verlsst. Hilft, die von Expat
     * unntigerweise aufgelsten XML-Entities wiederherzustellen.
     */
    var $cdata_active=false;
    /**
    * Constructor.
    *
    * @param string $file XML-Quelldatei
    * @param string $out optional Ausgabeformat (z.B. xml oder hixml)
    * @param string $outfile optional Ausgabedatei
    * @access public
    */
    function xml_line ($file, $out = "", $outfile = "")
    {
        $this->file = $file;
        if ($out) { // Augabe in csv oder xml mglich
            $this->set_pointer(); //Bei Dateiausgabe notwendig
            if (trim($out) == "csv") {
                $this->output_action = "csv";
            } else if (trim($out) == "hixml") {
                $this->output_action = "xml";
                $this->highlight = true;
            } else $this->output_action = "xml";
            if ($file) { // ohne file wird die Ausgabe im String gepuffert;
                $this->outfile = trim($outfile);
            }
        } else {
            // Hier wird weder eine Datei noch ein String ausgegeben.
            // Sinnvoll fr reine Abfragen.
            $this->output_action = false;
        }
    }

    function set_pointer()
    {
        $this->pointer_A = true;
        $this->pointer_D = true;
        $this->pointer_S = true;
        $this->pointer_E = true;
        $this->pointer_EOS = true;
        $this->pointer_BE = true;
    }

    /**
    * Setzt den Namen der Funktion, die auf den Stream zugreift
    *         und bergibt eine Parameterliste oder -string
    *
    * @param  $func string
    * @param  $arg mixed                (string or array)
    * @param  $command string
    * @param  $value mixed                (string or array)
    */
    function set_action($func, $arg = "", $command = "", $value = "")
    {
        // Alle Aktionen mssen durchgefhrt werden, wenn eine Ausgabe erfolgt
        // oder wenn get_record aktiv ist:
        if ($command == "get_record") {
            $this->set_pointer();
        } else {
            // Die Elementfunktionen brauchen D (Data)
            if ($func == "find_data") $this->pointer_D = true;
            // Die Attributfunktionen brauchen A (Attribut)
            else if ($func == "find_attribute") $this->pointer_A = true;
            // ==============================
            if ($command == "change_attribute_value"
                or $command == "delete_attribute"
                or $command == "insert_after_element"
                ) $this->pointer_E = true;

            else if ($command == "delete_element"
                or $command == "add_attributes"
                or $command == "replace_element"
                ) $this->pointer_S = true;

            else if ($command == "insert_element" // OK
                ) $this->pointer_EOS = true;

            else if ($command == "insert_before_end" // OK
                or $command == "delete_content"
                or $command == "replace_content"
                or $command == "get_content"
                or $command == "copy_branch"
                ) {
                    $this->set_pointer();
                    //$this->pointer_BE = true;
                }
        }
        // ================================================================
        $tmpindex = count($this->action_func); //Knftige Prozessnummer ermitteln
        // Datapart abtrennen falls vorhanden: 1-2-2:3
        if (isset($arg[4])) { // ist der Parameter vorhanden?
            if (strstr($arg[4], ':')) { // ist ein Trennzeichen vorhanden
                $tmparray = explode(':', $arg[4]);
                $arg[4] = $tmparray[0];

                $this->procdata[$tmpindex]['datapart'] = $tmparray[1];
            }
        }
        $this->action_func[] = array($func, $arg, $command, $value);
        $this->procdata[$tmpindex]['limitcount'] = 0;
    }

    /**
    * Motor des Ganzen. Wird bei jedem Neueintrag in $ls aufgerufen
    * und arbeitet die Funktionen ab, die im Hauptprogramm gesetzt wurden.
    * Jeder neue "Datensatz" lst die Aktionen aus.
    *
    * @access private
    */
    function xml_action()
    {
        $counter = 0;
        for ($i = 0; $i < count($this->action_func) ; $i++) {
            $this->func_counter = $i;
            // Funktionsname wird hier durch $vor dem f zur Funktion, die gleichzeitig aufgerufen wird.
            if ($this->action_func[$i]) { // wenn der Wert 0 ist, wurde die Funktion durch check_limit() gekillt,
                // weil keine Daten mehr gesucht oder verndert werden mssen.
                $this->{$this->action_func[$i][0]}($this->action_func[$i][1], $this->action_func[$i][2], $this->action_func[$i][3]);
                $counter++;
            }
        }
        if ($counter == 0) $this->actioncounter = false; //TEST
        if (! empty($this->copy_flag)) $this->copy_xml();
        if ($this->output_action) {
            if ($this->write_flag) {
                if ($this->output_action == "xml") $this->write_xml();
                else if ($this->output_action == "csv") $this->write_csv();
            } else {
                if ($this->last_nowrite_line == "") $this->write_flag = true;
                else if ($this->last_nowrite_line == $this->ls[0] . implode('-', $this->elc_path_array) . $this->ls[2]) {
                    $this->write_flag = true;
                    $this->last_nowrite_line = "";
                }
            }
            // Einfgen von Daten ===============================================
            // Replace-Daten oder Einschbe werden in einem Hash gespeichert,
            // der als Key die Reihe enthlt, hinter der eingefgt wird.
            $insertkey = $this->ls[0] . implode('-', $this->elc_path_array) . $this->ls[2];
            if (isset ($this->fill[$insertkey]) and $this->output_action == "xml") {
                $this->write_xml($this->fill[$insertkey]);
            }
        } //TESTEnde
    }

    /**
    * Startet den Parser und gibt die gefundenen Daten oder false zurck.
    *
    * @param  $tenc string optional    Targetencoding
    * @param  $enc string optional     Zeichensatz des Input-XML
    * @access public
    * @return array
    */
    function xml_stream($tenc="", $enc="")
    {
        switch ($enc) {
            case '':
               $enc=$this->encoding;
               break;
            case 'lat1':
            case 'latin1':
            case 'Latin1':
            case 'LATIN1':
            case 'l':
                $enc = 'ISO-8859-1';
                break;

            case 'utf8':
            case 'utf-8':
                $enc = 'UTF-8';
        }
        switch ($tenc) {
            case '':
               $targetenc=$this->targetencoding;
               break;
            case 'lat1':
            case 'latin1':
            case 'Latin1':
            case 'LATIN1':
            case 'l':
                $targetenc = 'ISO-8859-1';
                break;

            case 'utf8':
            case 'utf-8':
                $targetenc = 'UTF-8';
        }

        if ($this->outfile) {
            $this->of = fopen ($this->outfile, "w") or die("Fehler beim ffnen der Ausgabe");
        }

        $this->parser = xml_parser_create($enc);
        // Parser-Funktionen im Objekt:
        xml_set_object($this->parser, $this);
        xml_parser_set_option($this->parser, XML_OPTION_CASE_FOLDING, false);
        if ($targetenc) {
            xml_parser_set_option($this->parser, XML_OPTION_TARGET_ENCODING, $targetenc);
        }
        xml_set_default_handler($this->parser, "defaultData");
        xml_set_element_handler($this->parser, "startElement", "endElement");
        xml_set_character_data_handler($this->parser, "characterData");

        $string_flag = true; //Zeiger setzen fr Eingabeformat.
        // Ermitteln der Eingabe. String enthlt auf jeden Fall < Klammer.
        // Dateiname auf jeden Fall keine!
        $pos = strpos ($this->file, "<");
        if ($pos === false) { // Achtung: 3 Gleichheits-Zeichen
            $string_flag = false;
        }

        if (! $string_flag) {
            if (!($fp = fopen($this->file, "r"))) {
                die ("could not open XML input");
                // entweder es laufen noch Funktionen oder es gibt eine Ausgabe
            } while ($data = fread($fp, 4096) and ($this->actioncounter or $this->output_action)) {
                if (!xml_parse($this->parser, $data, feof($fp))) {
                    die (sprintf("XML error: %s at line %d",
                            xml_error_string(xml_get_error_code($this->parser)),
                            xml_get_current_line_number($this->parser)));
                }
            }
        } else {
            $data = $this->file;
            if (!xml_parse($this->parser, $data)) {
                die (sprintf("XML error: %s at line %d",
                        xml_error_string(xml_get_error_code($this->parser)),
                        xml_get_current_line_number($this->parser)));
            }
        }
        xml_parser_free($this->parser);
        // Ausgabe schlieen
        if ($this->outfile) {
            fclose($this->of) or die("Fehler beim Schlieen der Ausgabe!");
        }
        if (! $string_flag) { // Falls Eingabedatei gelesen wurde
            fclose($fp) or die("Fehler beim Schlieen der Eingabedatei!");
        }
        return $this->found_data;
    }

    /**
    * Sorgt fr die Ausgabe von Kommentaren und Doctype.
    *
    * @param integer $parser resource_id
    * @param string $ddata
    */
    function defaultData($parser, $ddata)
    {


        if ($ddata == '<![CDATA[') {

        	$this->cdata_active=true;

        } else if ($ddata == ']]>') {

        	$this->cdata_active=false;
        }

        // DefaultData direkt ausgeben ===
        if ($this->write_flag and $this->ignore_default_data == false) {
            // Vorhandene Daten schreiben ====================
            if ($this->data_buffer and $this->cdata_active == false) { // Wenn Daten im Puffer sind
                $this->ls[0] = "D"; //Data
                $this->ls[1] = $this->level;
                $this->ls[2] = $this->path_string(0);
                $this->ls[3] = "";
                $this->ls[4] = $this->data_buffer;
                $this->data_buffer = ""; //Reset
                if ($this->pointer_D) $this->xml_action();
            }
            if ($this->output_action == "xml") {
            	$this->write_xml($ddata);
            }
            else if ($this->output_action == "csv") {
                $this->write_csv("X;" . implode('-', $this->elc_path_array) . ";" . $this->path_string(0) . ";'';" . $ddata);
            }
        }
    }

    /**
    * Aktionen beim Auftauchen eines Starttags
    *
    * @param integer $parser resource
    * @param string $name
    * @param array $attrs
    * @access private
    */
    function startElement($parser, $name, $attrs)
    {

        if ($this->restore_xml_entities) {
            //Parser wandelt XML-Entities um. Dies wird hier wieder
            //rckgngig gemacht, falls der Schalter auf true steht.
            $attrs = $this->give_back_xml_entities($attrs);
        }
        // Vor dem Element knnen ebenfalls Daten sein :
        if ($this->level > 0) { // Nichts vor das Wurzelelement schreiben!
            if ($this->data_buffer) { // Wenn Daten im Puffer sind
                $this->ls[0] = "D"; //Data
                $this->ls[1] = $this->level;
                $this->ls[2] = $this->path_string(0);
                $this->ls[3] = "";
                $this->ls[4] = $this->data_buffer;
               // $this->data_buffer = ""; //Reset
                $this->data_buffer = $this->tmp_sign; //TEST
                if ($this->pointer_D) $this->xml_action();
            }
        }
        // Aktuelles Element global bekannt machen.
        $this->akt_name = $name;
        $this->level++;

        $this->ls[0] = "S"; //Starttag
        $this->ls[1] = $this->level;
        $this->ls[2] = $this->path_string(1);
        $this->ls[3] = "";
        $this->ls[4] = "";

         $this->temp_attr_buffer[$this->ls[2]] = $attrs;

        if ($this->pointer_S) $this->xml_action();

        if (! empty($this->fillattr)) {
            // Wenn Attribute hinzugefgt werden sollen, wird das hier gemacht:
            // Die Attribute sind im Hash gespeichert
            $aktattrbuffer = "S" . implode('-', $this->elc_path_array) . $this->path_string(0);


            if (isset($this->fillattr[$aktattrbuffer])) {

                // Attributhash anfgen
                $attrs = array_merge($attrs, $this->fillattr[$aktattrbuffer]);
                // Arrays zusammenfgen.
            }
        }
        foreach($attrs as $key => $value) {
            $this->ls[0] = "A"; //Attribut
            // $this->ls[1] = $this->level;
            $this->ls[2] = $this->path_string(0);
            $this->ls[3] = $key;
            $this->ls[4] = $value;
            //$this->ls[4] = $this->give_back_xml_entities($value);
           if ($this->pointer_A) $this->xml_action();
        }
        // Endmarke Starttag:
        $this->ls[0] = "EOS"; //Data
        //$this->ls[1] = $this->level;
        // $this->ls[2] = $this->path_string(0);
        $this->ls[3] = ""; // Attributwerte bleiben unbercksichtigt
        $this->ls[4] = ""; //
        if ($this->pointer_EOS) $this->xml_action();
    }
    /**
     * Von Expat umgewandelte XML-Entities wieder zurckverwandeln
     */
    function give_back_xml_entities($attrs) {

    	$back_arr=array();
        foreach ($attrs as $k => $v) {
        	$back_arr[$k] = $this->replace_xml_entities($v);
        }
    	return $back_arr;
    }
    function replace_xml_entities($val) {
     	//$se = array ('&', '<', '>', '"', "'");
    	//$re = array('&amp;', '&lt;', '&gt;', '&quot;', '&apos;');
    	//return (str_replace($se, $re, $val));
    	return (htmlspecialchars($val, ENT_QUOTES));
    }

    /**
    * Standard-XML-Parser-Funktion
    *
    * @param integer $parser resource
    * @param string $name
    * @access private
    */
    function endElement($parser, $name)
    {
        // Jedes Element, das sonst leer wre, oder nur Whitespace enthlt, bekommt
        // temporr eine Kennung $this->tmp_sign.
        if ($this->ls[4] == "" and $this->data_buffer == "") {
            $this->data_buffer = $this->tmp_sign . $this->ls[4];
        }

        $this->ls[0] = "D"; //Data
        $this->ls[1] = $this->level;
        $this->ls[2] = $this->path_string(0);
        $this->ls[3] = "";
        // Wenn das Element geschlossen wurde, werden die gesammelten Daten geschrieben.
        $this->ls[4] = $this->data_buffer;
        if ($this->pointer_D) $this->xml_action();
        // Fr Einfgungen hinter dem Endtag
        $this->ls[0] = "BE"; //Before End
        // $this->ls[1] = $this->level;
        // $this->ls[2] = $this->path_string(0);
        // $this->ls[3] = "";
        // Wenn das Element geschlossen wurde, werden die gesammelten Daten geschrieben.
        $this->ls[4] = "";
        if ($this->pointer_BE) $this->xml_action();
        $this->ls[0] = "E"; //End
        $this->ls[1] = $this->level;
        $this->ls[2] = $this->path_string(-1);
        $this->ls[3] = "";
        $this->ls[4] = "";
        if ($this->pointer_E) $this->xml_action();
        $this->level--;
        array_pop($this->elc_path_array); // Darf erst nach der Action gekrzt werden!!!
        $this->data_buffer = ""; //Reset
    }

    /**
    * Standard-XML-Parser-Funktion
    *
    * @param integer $parser resource
    * @param string $data
    * @access private
    */
    function characterData($parser, $data)
    {


    	//=== Test : Nur auerhalb von CDATA-Abschnitten und wenn eingeschaltet
        if (! $this->cdata_active and $this->restore_xml_entities) {
            /*
            $re = array ('&', '<', '>', '"', "'");
    	    $se = array('&amp;', '&lt;', '&gt;', '&quot;', '&apos;');
    	    $data = str_replace($se, $re, $data);
            */
            $data = $this->replace_xml_entities($data);
    	}
        //====

        // $data wird in mehreren Teilen ausgegeben. Daher wird hier gesammelt.
        $this->data_buffer .= $data; //Daten sammeln und zusammenfhren
    }

    /**
    * Level-Counter-Funktion
    * Registriert die Ebene des XML-Elements und aktualisiert den "Pfad" des
    * aktuellen Elements im $path_array. Liefert einen String aus allen Werten
    * des $path_array zurck
    * Gleichzeitig werden alle gleichnamigen Elemente gezhlt, solange sie sich innerhalb
    * eines bergeordeten Elements befinden.
    *
    * @param integer $plusminus
    * @return string
    * @access private
    */
    function path_string($plusminus)
    {
        if ($plusminus == 1) {
            $this->path_array[] = $this->akt_name;
            $tmp = join('/', $this->path_array);
            // element-level-counter
            if (!isset($this->elc[count($this->path_array)][$this->akt_name])) {
                $this->elc[count($this->path_array)][$this->akt_name] = 1;
            } else {
                $this->elc[count($this->path_array)][$this->akt_name]++;
            }
            // array enthlt nur die aktuellen Zhlerwerte:
            $this->elc_path_array[] = $this->elc[count($this->path_array)][$this->akt_name];
        } else if ($plusminus == -1) {
            // Zhler fr Elemente innerhalb des Elements abschlieen.
            $this->elc[count($this->path_array) + 1] = array();
            // Datapart-counter abschlieen:
            $tmp = join('/', $this->path_array); //Wert erst nach Rckgabe herunterzhlen
            array_pop($this->path_array);
        } else {
            $tmp = join('/', $this->path_array);
        }
        return $tmp;
    }

    /**
    * Setzt die Abfrage fr die Werte des Attributs oder der Attribute, die der Vorgabe
    * entsprechen. Ohne Parameter werden alle Werte in das Ergebnis-Array geschrieben.
    *
    * @param integer $limit optional  0 bedeutet: kein Limit
    * @param string $path optional
    * @param string $pattern optional
    * @param string $attrn optional
    * @param string $elcount optional
    * @access public
    */
    function get_attribute($limit = 0, $path = "", $pattern = "", $attrn = "", $elcount = "")
    {
        $arg = array($limit, $path, $pattern, $attrn, $elcount);
        $this->set_action("find_attribute", $arg);
    }

    /**
    * Benennt das angegebene Attribut um.
    *
    * @param array $arg
    * @param string $value
    * @access public
    */
    function rename_attribute ($arg, $value)
    {
        $this->set_action("find_attribute", $arg, "rename_attribute", $value);
    }

    /**
    * Setzt eine nderungsabfrage. Funktioniert hnlich wie get_attribute. Die dort bergebenen
    * Parameter werden hier als Array bergeben. Als zweiter Parameter kommt der String oder
    * die Zahl, die den alten Wert ersetzt. Die Funktion wird sooft angewendet, wie sie auf
    * die Abfrage trifft. Der entfernte Wert wird nicht gespeichert.
    *
    * @param array $arg
    * @param mixed $value (string or integer)
    * @access public
    * @see get_attribute
    */
    function change_attribute_value($arg, $value)
    {
        $this->set_action("find_attribute", $arg, "change_attribute", $value);
    }

    /**
    * Lscht das angegebene Attribut.
    *
    * @param mixed $limit optional  0 bedeutet: kein Limit. Neu in 0.4.3.6:
    *                               alle Parameter knnen auch als Array bergeben werden.
    * @param string $path optional
    * @param string $pattern optional
    * @param string $attrn optional
    * @param string $elcount optional
    * @access public
    */
    function delete_attribute($limit = 0, $path = "", $pattern = "", $attrn = "", $elcount = "")
    {
        if (is_array($limit)) {
            $arg = $limit;
        } else {
            $arg = array($limit, $path, $pattern, $attrn, $elcount);
        }
        $this->set_action("find_attribute", $arg, "delete_attribute");
    }

    /**
    * Ersetzt das angegebene Attribut durch ein anderes.
    *
    * @param array $arg
    * @param array $newattr
    * @see get_attribute
    * @access public
    */
    function replace_attribute($arg, $newattr)
    {
        $this->set_action("find_attribute", $arg, "replace_attribute", $newattr);
    }

    /**
    * Eigentliche Funktion fr die Verarbeitung der Abfragen.
    * bernimmt das Sucharray, optional den Namen der Funktion zur Modifizierung der Aktion
    * und eventuell einen neuen Wert bzw. Werte.
    *
    * @param array $arg
    * @param string $command optional
    * @param mixed $newvalue optional (string or array)
    * @access private
    */
    function find_attribute ($arg, $command = "", $newvalue = "")
    {
        if ($this->ls[0] != "A") return;

        $limit = (isset($arg[0])) ? $arg[0] : 0;
        $path = (isset($arg[1])) ? $arg[1] : "";
        $pattern = (isset($arg[2])) ? $arg[2] : "";
        $attr = (isset($arg[3])) ? $arg[3] : "";
        // Attribut kommt hier als String!
        $elcount = (isset($arg[4])) ? $arg[4] : "";
        $elementname = array_pop(explode("/", $this->ls[2]));
        // Sucht nur nach dem Schluss von $path, wenn nur der Elementname
        // angegeben wird.
        $tmp_path = $this->wildcardpath($path);
        if ($path == "" or $path == $tmp_path or $path == array_pop(explode("/", $this->ls[2]))) {
            // Attribute und Werte mit Suchmuster vergleichen
            if ($attr == "" or $attr == $this->ls[3]) {
                if ($elcount == "" or $this->elc_check($elcount) == true) {
                    if ($pattern == "" or $pattern == $this->ls[4]) { // genaue Suche fr den Attributwert!
                        if ($command == "") {
                            // $elementname = $this->path_array[count($this->path_array) - 1];
                            $this->found_data[$this->func_counter][] = array($this->ls[4], $elementname, $this->ls[3], $this->ls[2], implode('-', $this->elc_path_array));
                            $this->check_limit($this->func_counter, $limit); //Checken, ob die Funktion abrechen kann

                        } else if ($command == "change_attribute") $this->ls[4] = $newvalue;
                        else if ($command == "rename_attribute") $this->ls[3] = $newvalue;
                        else if ($command == "delete_attribute") $this->write_flag = false;
                        else if ($command == "replace_attribute") {
                            list($this->ls[3], $this->ls[4]) = each ($newvalue);
                        }
                    }
                }
            }
        }
    }

    /**
    * Ruft set_action auf und bergibt das Kommando "find_data" und die Suchkriterien als Parameter.
    * set_action wird bei fast jedem Event des Expat-Parsers aufgerufen, und ruft wiederum die
    * funktion find_data (und die brigen gesetzen Funktionen) auf.
    *
    * @param integer $limit optional  Anzahl der gesuchten Treffer. 0 = keine Einschrnkung
    * @param string $path optional  root/element1/element2   (kann unvollstndig sein, z.B. element2)
    * @param string $pattern optional  String, der in den gesuchten Daten enthalten sein muss
    * @param mixed $attrn optional  array(name=>value, name2=>value2) Attribut(e) mssen enthalten sein.
    * @param string $elcount optional  "1-6-2" oder "1-4-6:2" oder "1-6-{3 to 5}" oder "1-6-{3,5,8}"
    * @access public
    */
    function get_data ($limit = 0, $path = "", $pattern = "", $attrn = "", $elcount = "")
    {
        $this->set_action("find_data", array($limit, $path, $pattern, $attrn, $elcount));
    }

    /**
    * bergibt das Kommando, alle Elemente innerhalb dieses Elementes als Ergebnis zurckzuliefern.
    * Alle enthaltenen Elemente werden geliefert, in dem ein Zeiger gesetzt wird, der erst wieder
    * entfernt wird, wenn das Schlusstag gefunden wurde.
    *
    * @param integer $limit
    * @param string $path
    * @param string $pattern
    * @param mixed $attrn
    * @param string $elcount
    * @see get_data
    * @access public
    */
    function get_record ($limit = 0, $path = "", $pattern = "", $attrn = "", $elcount = "")
    {
        //Erlaubt das Setzen eines Levels. Pfade werden ignoriert.
        if ($elcount) {
        	$this->procdata[$this->func_counter]['maxlevel'] = $elcount;
        	$elcount = '';
        }
        $this->set_action("find_data", array($limit, $path, $pattern, $attrn, $elcount), "get_record");
    }

    /**
    * Ersetzt den gefundenen String durch einen neuen.
    * Zustzlich zur Parameterliste $arg wird ein Wert bergeben, der den gefundenen String ersetzt.
    *
    * @param array $arg
    * @param string $value
    * @see get_data
    * @access public
    */
    function change_data ($arg, $value)
    {
        $this->set_action("find_data", $arg, "change_data", $value);
    }

    /**
    * Hngt den neuen Inhalt an den Wert an
    * Zustzlich zur Parameterliste $arg wird ein Wert bergeben, der den gefundenen String erweitert.
    *
    * @param array $arg
    * @param string $value
    * @see get_data
    * @access public
    */
    function append_data ($arg, $value)
    {
        $this->set_action("find_data", $arg, "append_data", $value);
    }

    /**
    * Lscht das gesuchte Element und seinen Inhalt, indem die Ausgabe unterbrochen wird.
    *
    * @param integer $limit optional  Anzahl der gesuchten Treffer. Danach wird die Funktion beendet
    * @param string $path optional  root/element1/element2   (kann unvollstndig sein, z.B. element2)
    * @param string $pattern optional  String, der in den gesuchten Daten enthalten sein muss
    * @param mixed $attrn optional  array(name=>value, name2=>value2) Attribut(e) mssen enthalten sein.
    * @param string $elcount optional  "1-6-2" oder "1-4-6:2" oder "1-6-{3 to 5}" oder "1-6-{3,5,8}"
    * @access public
    * @see get_attribute
    */
    function delete_element ($limit = 0, $path = "", $pattern = "", $attrn = "", $elcount = "")
    {
        $this->set_action("find_data", array($limit, $path, $pattern, $attrn, $elcount), "delete_element");
    }

    /**
    * Ist identisch mit delete_element. Zustzlich werden Daten in einen Speicher (fill)
    * geschrieben und dieser beim Erreichen der Marke in die Ausgabe eingefgt.
    *
    * @param array $arg
    * @param string $replace
    * @access public
    * @see get_data
    */
    function replace_element ($arg, $replace)
    {
        $this->set_action("find_data", $arg, "delete_element", $replace);
    }

    /**
    * Lscht den Inhalt des gesuchten Elements, indem die Ausgabe unterbrochen wird.
    *
    * @param integer $limit optional  Anzahl der gesuchten Treffer. Danach wird die Funktion beendet
    * @param string $path optional  root/element1/element2   (kann unvollstndig sein, z.B. element2)
    * @param string $pattern optional  String, der in den gesuchten Daten enthalten sein muss
    * @param mixed $attrn optional  array(name=>value, name2=>value2) Attribut(e) mssen enthalten sein.
    * @param string $elcount optional  "1-6-2" oder "1-4-6:2" oder "1-6-{3 to 5}" oder "1-6-{3,5,8}"
    * @access public
    * @see get_attribute
    */
    function delete_content ($limit = 0, $path = "", $pattern = "", $attrn = "", $elcount = "")
    {
        $this->set_action("find_data", array($limit, $path, $pattern, $attrn, $elcount), "delete_content");
    }
    // neu ============================================
    function get_content ($limit = 0, $path = "", $pattern = "", $attrn = "", $elcount = "")
    {
        $this->set_action("find_data", array($limit, $path, $pattern, $attrn, $elcount), "get_content");
    }
    // gleiche Funktion wie oben. Replace-Parameter wird als flag missbraucht.
    // "CB" signalisiert, dass am Ende die Start- und Endtag eingefgt werden.
    function copy_branch ($limit = 0, $path = "", $pattern = "", $attrn = "", $elcount = "")
    {
        $this->set_action("find_data", array($limit, $path, $pattern, $attrn, $elcount), "get_content", "CB");
    }
    // ============ neu Ende ================================
    function rename_element ($arg, $replace)
    {
        $this->set_action("find_data", $arg, "rename_element", $replace);
    }

    /**
    * == replace_content
    * Ist identisch mit delete_content. Zustzlich werden Daten in einen Speicher (fill)
    * geschrieben und dieser beim Erreichen der Marke in die Ausgabe eingefgt.
    *
    * @param array $arg
    * @param string $replace
    * @access public
    * @see get_data
    */
    function replace_content ($arg, $replace)
    {
        $this->set_action("find_data", $arg, "delete_content", $replace);
    }

    /**
    * Fgt Element(e) oder XML-Code an der gesuchten Stelle ein.
    * Der Inhalt wird direkt hinter dem Start-Tag eingesetzt.
    * Man kann dadurch auch den Zwischenraum zwischen zwei Starttags fllen, selbt wenn
    * dort kein Leerzeichen ist. <p><b>wort</b>text</p>. Eingefgt wird so:
    * <p>neuer Text<b>wort</b>text</p>
    * Es knnen auch leere Elemente gefllt werden: <textarea></textarea>
    * <textarea>Text</textarea>
    *
    * @param array $arg
    * @param string $insert
    * @access public
    * @see get_data
    */
    function insert_element ($arg, $insert)
    {
        $this->set_action("find_data", $arg, "insert_element", $insert);
    }

    /**
    * Fgt Attribute in einem Hash hinzu. Nutzt ebenfalls die Startelementmarke und nicht die Attributmarke,
    * weil direkt auf das Expat-Attributarray zugegriffen wird, bevor dieses weiter verarbeitet wird.
    * (Das muss man sich klar machen, wenn zunchst Attribute zugefgt und gleichzeitig nach gleichnamigen
    * Attributen gesucht wird!)
    *
    * @param array $arg
    * @param string $insert
    * @access public
    * @see get_data
    */
    function add_attributes ($arg, $insert)
    {
        $this->set_action("find_data", $arg, "add_attributes", $insert);
    }

    /**
    * Fgt XML-Code oder einen String hinter dem End-Tag eines Elements ein.
    *
    * @param array $arg
    * @param string $insert
    * @access public
    * @see get_data
    */
    function insert_after_element ($arg, $insert)
    {
        $this->set_action("find_data", $arg, "insert_after_element", $insert);
    }

    /**
    * Fgt einen String vor dem End-Tag eines Elements ein.
    *
    * @param array $arg
    * @param string $insert
    * @access public
    * @see get_data
    */
    function insert_before_end ($arg, $insert)
    {
        $this->set_action("find_data", $arg, "insert_before_end", $insert);
    }

    /**
    * Findet und liefert den Elementinhalt mit einem bestimmten id-Attribut,
    * z.B. find_data_by_id(15); <Kiste id="1">Schrauben</Kiste>
    *
    * Shorthand fr get_data(). Heit "id" nicht "id" sondern "nr", kann der 2. Parameter
    * gesetzt werden. Damit wird aus der Funktion eigentlich ein find_data_by_attribute_name_value.
    *
    * @param string $value
    * @param string $idname optional (anderer Atrributname, wenn nicht "id")
    * @access public
    */
    function find_data_by_id($value, $idname = "id")
    {
        $this->set_action("find_data", array(0, "", "", array($idname => $value)));
    }

    /**
    * Ersetzt den Elementwert eines Elements mit einem bestimmten id-Attribut,
    * z.B. change_data_by_id(15,"Ngel"); <Kiste id="1">Schrauben</Kiste>
    *
    * @param string $value
    * @param string $replace
    * @param string $idname optional (anderer Atrributname, wenn nicht "id")
    * @access public
    * @see find_data_by_id
    */
    function change_data_by_id($value, $replace, $idname = "id")
    {
        $this->set_action("find_data", array(0, "", "", array($idname => $value)), "change_data", $replace);
    }

    /**
    * Interne Hauptfunktion, die von den oben beschriebenen "Pseudofunktionen" in die Schleife gesetzt wird.
    *
    * @param array $arg bekannte Kriterienliste
    * @param string $command optional  Name der aufrufenden Funktion, damit differenziert werden kann
    * @param string $changedata optional  Ein Ersetzugs- oder Einsetzungsstring
    * @access private
    */
    function find_data($arg, $command = "", $changedata = "")
    {
        $saveall = false;
        if ($this->write_flag != true) return;

        if ($command == "rename_element") {
            // kann eventuell auf S und E reduziert werden, fr csv aber notwendig, dass
            // alle Elementpfade aktualisiert werden.;
        } else if ($this->ls[0] == "D") {
            switch ($command) {
                case "":
                    break;
                case "insert_after_element":
                    break;
                case "change_data":
                    break;
                case "get_record":
                    if (isset($this->procdata[$this->func_counter]['maxlevel'])) {
	                    if ($this->level > $this->procdata[$this->func_counter]['maxlevel']) {
	                    	break;
	                    }
                    }
                    if (isset($this->procdata[$this->func_counter]['recordmode'])) {
                        $saveall = true;
                        if (isset($this->procdata[$this->func_counter]['rowcounter'])) {
                            $rc = $this->procdata[$this->func_counter]['rowcounter'];
                        } else {
                            $this->procdata[$this->func_counter]['rowcounter'] = 0;
                            $rc = 0; //rowcounter
                            $this->procdata[$this->func_counter]['outerattr'] = 1;
                        }
                        /*  War Buggy, fhrte zu doppelten Attributeintrgen als _@   */
                        if (isset($this->temp_attr_buffer[$this->ls[2]]) and ! empty($this->procdata[$this->func_counter]['outerattr'])) {
                            foreach ($this->temp_attr_buffer[$this->ls[2]] as $k => $v) {
                                $this->table_result[$this->func_counter][$rc]['_@' . $k] = $v;
                            }
                            $this->procdata[$this->func_counter]['outerattr'] = 0;
                        }
                    }
                    break;
                case "append_data":
                    break;
                default:
                    return;
            }
        } else if ($this->ls[0] == "A") return;

        else if ($this->ls[0] == "S") {
            switch ($command) {
                case "delete_element":
                    break;
                case "rename_element":
                    break;
                case "add_attributes":
                    break;
                case "get_record":
                    break;
                default:
                    return;
            }
        } else if ($this->ls[0] == "E") {
            switch ($command) {
                // case "insert_after_element":
                // break;
                case "rename_element":
                    break;
                case "get_record":
                    break;
                default:
                    return;
            }
        } else if ($this->ls[0] == "EOS") {
            switch ($command) {
                case "delete_content":
                    break;
                case "get_content":
                    break;
                case "insert_element":
                    break;
                case "get_record":
                    break;
                default:
                    return;
            }
        } else if ($this->ls[0] == "BE") {
            switch ($command) {
                case "insert_before_end":
                    break;
                case "get_content":
                    break;
                case "get_record":
                    break;
                default:
                    return;
            }
        }

        $limit = (isset($arg[0])) ? $arg[0] : 0;
        $path = (isset($arg[1])) ? $arg[1] : "";
        $pattern = (isset($arg[2])) ? $arg[2] : "";
        $attr = (isset($arg[3])) ? $arg[3] : "";
        // Attribute kommen als Hash name=>value rein
        $elcount = (isset($arg[4])) ? $arg[4] : "";
        $elementname = array_pop(explode("/", $this->ls[2]));
        $levelsig = implode('-', $this->elc_path_array);

        if ($attr and !$saveall) {
            $attrcheck = false;
            $counter = count($attr); //Der Counter wird hochgesetzt auf die Attributzahl
            foreach ($attr as $k => $v) {
                // Der Attributpuffer enthlt die Werte zu den gesuchten Daten.
                if (isset($this->temp_attr_buffer[$this->ls[2]][$k])) {
                    if ($this->temp_attr_buffer[$this->ls[2]][$k] == $v) {
                        $counter--; //Counter runterzhlen.
                    }
                }
            }
            if (!$counter) {
                $attrcheck = true; //Wenn alle Attribute und Werte mit der Vorgabe
            } //berienstimmen, ist der Counter heruntergezhlt.
            else return; //Braucht nicht weiter verfolgt zu werden
        }

        $tmp_path = $this->wildcardpath($path);

        if ($saveall or $path == "" or $path == $tmp_path or $path == array_pop(explode("/", $this->ls[2]))) {
            if ($saveall or $attr == "" or $attrcheck == true) {
                // Teilsuche
                if ($saveall or $pattern == "" or strstr($this->ls[4], $pattern)) {
                    if ($saveall or $elcount == "" or $this->elc_check($elcount) == true) {
                        // Datapart:
                        if ($saveall or !isset($this->procdata[$this->func_counter]['datapart'])) {
                            $part = true;
                        } else {
                            // $levelsig = implode('-', $this->elc_path_array);
                            if (!isset($this->found_data_part[$this->func_counter . $this->ls[2] . $levelsig])) {
                                $this->found_data_part[$this->func_counter . $this->ls[2] . $levelsig] = 1;
                            } else {
                                $this->found_data_part[$this->func_counter . $this->ls[2] . $levelsig]++;
                            }
                            if ($this->procdata[$this->func_counter]['datapart'] == $this->found_data_part[$this->func_counter . $this->ls[2] . $levelsig]) {
                                $part = true;
                            } else {
                                $part = false;
                            }
                        }

                        if ($saveall or $part == true) {
                            // $levelsig = implode('-', $this->elc_path_array);
                            if ($command == "") {
                                // $elementname = $this->path_array[count($this->path_array) - 1];
                                // Erst sehen, ob Attribute da sind, wenn nicht leeres Array bergeben.
                                $tmpattrb = (isset($this->temp_attr_buffer[$this->ls[2]])) ? $this->temp_attr_buffer[$this->ls[2]] : array();
                                $this->found_data[$this->func_counter][] = array(str_replace($this->tmp_sign, '', $this->ls[4]), $elementname, $tmpattrb, $this->ls[2], implode('-', $this->elc_path_array));
                                $this->check_limit($this->func_counter, $limit); //Limit-Checkfunktion aufrufen

                            } else if ($saveall == true and $this->procdata[$this->func_counter]['recordmode'] != $this->ls[2]) {
                                // nur Elemente != Wurzelelement des Datensatzes
                                if (trim($this->ls[4]) != "" or (isset($this->temp_attr_buffer[$this->ls[2]]) and count($this->temp_attr_buffer[$this->ls[2]]))) {
                                    // der vollstndige Suchpfad vom get_record-modus wird vom aktuellen Pfad abgezogen.
                                    // brig bleibt das aktuelle Element oder ein Teilstring, falls es sich um ein
                                    // Child-Element handelt.
                                    $en = str_replace($this->procdata[$this->func_counter]['recordmode'] . '/', '', $this->ls[2]);

                                    //$data = str_replace($this->tmp_sign, '', trim($this->ls[4]));
                                    $data = trim(str_replace($this->tmp_sign, '', $this->ls[4]));
                                    if (!isset($this->table_result[$this->func_counter][$rc][$en])) {
                                        $this->table_result[$this->func_counter][$rc][$en] = $data;
                                     // Attribute in Tabelle aufnehmen:
                                        if (count($this->temp_attr_buffer[$this->ls[2]])) {
                                           $this->table_result[$this->func_counter][$rc][$en . '@'] = $this->temp_attr_buffer[$this->ls[2]];
                                           foreach ($this->temp_attr_buffer[$this->ls[2]] as $k => $v) {
                                           	   $this->table_result[$this->func_counter][$rc][$en . '@'. $k] = $v;
                                           }
                                        }
                                    } else if ($data != "" or !empty($this->temp_attr_buffer[$this->ls[2]])) {
                                        //$this->table_result[$this->func_counter][$rc][$en] .= "!!!" . $data;
                                        // Ist schon ein Wert im Ergebnis-Hash, wird der Wert angehngt, aber nur,
                                        // wenn er kein Leerstring ist.
                                        if (is_string($this->table_result[$this->func_counter][$rc][$en])) {
                                        	if (isset($this->table_result[$this->func_counter][$rc][$en . '@'])) {
                                                $this->table_result[$this->func_counter][$rc][$en] = array(array('#' => $this->table_result[$this->func_counter][$rc][$en],
                                                                                                     '@' => $this->table_result[$this->func_counter][$rc][$en . '@']));
                                                 foreach ($this->table_result[$this->func_counter][$rc][$en . '@'] as $k => $v) {
                                                 	  unset($this->table_result[$this->func_counter][$rc][$en . '@' . $k]);
                                                 }

                                                 unset($this->table_result[$this->func_counter][$rc][$en . '@']);
                                        	} else {
                                        		$this->table_result[$this->func_counter][$rc][$en] = array(array('#' => $this->table_result[$this->func_counter][$rc][$en]));
                                        	}
                                        }
                                        if (!empty($this->temp_attr_buffer[$this->ls[2]])) {
                                        $this->table_result[$this->func_counter][$rc][$en][] = array('#' => $data,
                                                                                                     '@' => $this->temp_attr_buffer[$this->ls[2]]);

                                        } else {
                                        	$this->table_result[$this->func_counter][$rc][$en][] = array('#' => $data);
                                        }
                                    }

                                }
                            } else if ($command == 'get_record' and $this->ls[0] == "S") {
                                // Setzt die Marke aus Element- und Zhlpfad, damit alle Ergebnisse des
                                // Datensatzes registriert werden:
                                $this->procdata[$this->func_counter]['recordmode'] = $this->ls[2];
                            } else if ($command == 'get_record' and $this->ls[0] == "E") {
                                // Recordmode killen -Todo: hier knnte jetzt das Attributarray gelscht werden
                                unset($this->procdata[$this->func_counter]['recordmode']);
                                $this->procdata[$this->func_counter]['rowcounter']++;
                                // Flag setzen, damit die Attribute des umschlieenden Elements
                                // ebenfalls geschrieben werden.
                                $this->procdata[$this->func_counter]['outerattr'] = 1;
                                // hochzhlen fr nchsten Datensatz
                            } else if ($command == "change_data") {
                                $this->ls[4] = $changedata; //Setzt die bergebenen Daten ein.
                            } else if ($command == "append_data") {
                                $this->ls[4] .= $changedata; //hngt die bergebenen Daten an.
                            } else if ($command == "delete_element") {
                                $this->write_flag = false;
                                // Endmarke fr den Nowrite-Vorgang definieren:
                                $this->last_nowrite_line = "E" . $levelsig . $this->ls[2];
                                // Wenn das Element ersetzt wird, passiert das nach der letzten Zeile
                                if ($changedata) $this->fill[$this->last_nowrite_line] = $changedata;
                            } else if ($command == "delete_content") {
                                $this->write_flag = false;
                                // Endmarke fr den Nowrite-Vorgang definieren:
                                $this->last_nowrite_line = "BE" . $levelsig . $this->ls[2];
                                // Klammer wird nicht geschlossen, daher muss sie wieder eingefgt werden.
                                // if ($changedata) $this->fill[$this->last_nowrite_line] = '>' . $changedata;
                                // Klammer muss nachgetragen werden.
                                $this->fill[$this->last_nowrite_line] = '>' . $changedata;
                                // if ($changedata) $this->fill[$this->last_nowrite_line] = $changedata;
                            } else if ($command == "get_content") {
                                if ($this->ls[0] == "EOS") {
                                    // Marke setzen, damit content in buffer geschrieben wird.
                                    // Allgemeines Flag, um unntige Arrayzugriffe zu vermeiden
                                    $this->copy_flag[$this->func_counter] = "";
                                    // Buffer fr die Daten
                                    // Eventuell schon einmal das Ergebnisarray schreiben und gepufferten String
                                    // erst spter nachtragen
                                } else if ($this->ls[0] == "BE") {
                                    // erste berflssige Klammer von EOS wieder entfernen
                                    $copy_string = str_replace($this->tmp_sign, '', substr($this->copy_flag[$this->func_counter], 1));
                                    if ($changedata == "CB") {
                                        // Elementtemplate erzeugen und Inhalt hineinpacken:
                                        $copy_string = sprintf($this->make_element($elementname, $this->temp_attr_buffer[$this->ls[2]]), $copy_string);
                                    }
                                    $this->found_data[$this->func_counter][] = array($copy_string, $elementname, $this->temp_attr_buffer[$this->ls[2]], $this->ls[2], implode('-', $this->elc_path_array));
                                    unset($this->copy_flag[$this->func_counter]);
                                }
                            } else if ($command == "insert_element") {
                                $insertline = "EOS" . $levelsig . $this->ls[2];
                                if ($changedata) $this->fill[$insertline] = $changedata;
                            } else if ($command == "insert_before_end") {
                                $insertline = "BE" . $levelsig . $this->ls[2];
                                if ($changedata) $this->fill[$insertline] = $changedata;
                            } else if ($command == "insert_after_element") {
                                $insertline = "E" . $levelsig . $this->ls[2];
                                if ($changedata) $this->fill[$insertline] = $changedata;
                            }
                            // ========== TEST ====================
                            else if ($command == "rename_element" and $changedata) {
                                  //==================================
                                $tmparr = explode("/", $this->ls[2]);
                                $tmparr[count($tmparr)-1] = $changedata;
                                $this->ls[2] = join('/', $tmparr);
                                // =====================================
                            } else if ($command == "add_attributes") {
                                // Attribute werden als array bergeben
                                // Kennung und pathstring reichen als Key!
                                $insertline = "S" . $levelsig . $this->ls[2];
                                if ($changedata) $this->fillattr[$insertline] = $changedata;
                            }
                        }
                    }
                }
            }
        }
    }
    // ===========neu
    function make_element($eln, $attr_arr)
    {
        $attr_tpl = ' %s="%s"';
        $attr_str = "";
        // Attributstring zusammen schrauben:
        foreach ($attr_arr as $k => $v) {
            $attr_str .= sprintf($attr_tpl, $k, $v);
        }
        // Template zurckliefern:
        return '<' . $eln . $attr_str . '>%s</' . $eln . '>';
    }

    /**
    * Prft, ob das vorgegebene Limit der Suchergebnisse erreicht ist. Ist das der Fall,
    * wird die Funktion nicht mehr gebraucht.
    *
    * @param integer $funcnr Funktionsnummer
    * @param integer $lim vorgegebenes Limit
    * @access private
    * @see find_data
    */
    function check_limit($funcnr, $lim)
    {
        $this->procdata[$funcnr]['limitcount']++;
        if ($this->procdata[$funcnr]['limitcount'] == $lim) {
            $this->action_func[$funcnr] = 0; //gesamtes Funktionsarray mit 0 berschreiben.
        }
    }

    /**
    * Um einen Wildcardstring mit dem aktuellen pathstring vergleichen zu knnen, wird
    * hier in den pathstring an entsprechender Stelle ebenfalls ein * eingefgt.
    * gibt es kein * wird normal verglichen.
    *
    * @param string $path
    * @access private
    * @see find_data
    */
    function wildcardpath ($path)
    {
        if (strstr($path, '*')) { // wildcards im Pfad
            $tmp_path_array = explode('/', $path);
            $tmp_path = "";
            for ($i = 0; $i < count($this->path_array); $i++) {
                if (!isset($tmp_path_array[$i])) break; //Wenn der Index nicht mehr existiert
                // abbrechen
                $tmp_path .= ($tmp_path_array[$i] == '*') ? '*/' : $this->path_array[$i] . '/';
            }
            return substr ($tmp_path, 0, -1); //letzten Slash wieder entfernen

        } else {
            return $this->ls[2];
        }
    }

    /**
    * Hilfsfuktion: vergleicht den gesuchten element-counter-string mit dem aktuellen
    *
    * @param string $search_elcstring
    * @return boolean
    * @access private
    */
    function elc_check($search_elcstring)
    {
        if (! strstr($search_elcstring, '{')) {
            // direkter Stringvergleich
            $act_string = join ("-", $this->elc_path_array);
            if ($search_elcstring == $act_string) {
                return true;
            }
        } else if (! strstr($search_elcstring, '-')) {
            // nur ein Wert 1. Level aber unterschiedlich, sonst htte es oben gepasst
            return false;
        } else if (count(explode('-', $search_elcstring)) != count($this->elc_path_array)) {
            // arrays haben unterschiedliche Lnge, knnen also nicht passen.
            return false;
        } else {
            // Prfung der Einzelwerte
            $search_array = explode("-", $search_elcstring);
            for ($i = 0; $i < count($search_array); $i++) {
                if (($search_array[$i] != $this->elc_path_array[$i]) and (strstr($search_array[$i], '}') == false)) {
                    return false;
                } else if (strstr($search_array[$i], '*')) {
                    // wildcard *;
                } else if (strstr($search_array[$i], ',')) {
                    $search_array[$i] = str_replace ('{', '', $search_array[$i]);
                    $search_array[$i] = str_replace ('}', '', $search_array[$i]);
                    $temparray = explode(',', $search_array[$i]);
                    if (! in_array($this->elc_path_array[$i], $temparray)) {
                        return false;
                    }
                } else if (preg_match("/(\d+)\s*to\s*(\d+)/i", $search_array[$i], $matches)) {
                    if ($this->elc_path_array[$i] < $matches[1] or $this->elc_path_array[$i] > $matches[2]) {
                        return false;
                    }
                }
            }
            return true;
        }
    }

    /**
    * Experimentelle Funktion, die vielleicht wieder verschwindet. Mit ihr kann der aktuelle
    * "Linestring" als csv ausgegeben werden (buffer oder Datei). Dies ermglicht z.B. eine
    * komprimierte Darstellung
    * einer XML-Datei, die in jeder Zeile ein Event hat. Das knnte fr umfangreiche Arbeiten mit
    * groen Dateien interessant sein.
    *
    * @param string $linestring optional   Daten die eingefgt werden
    * @access private
    */
    function write_csv($linestring = "")
    {
        if (!$linestring) {
            $linestring .= $this->ls[0] . ";";
            // $linestring .= $this->ls[1] . ";";
            // kompletter Zhlpfad:
            $linestring .= implode('-', $this->elc_path_array) . ";";

            $linestring .= $this->ls[2] . ";";
            $linestring .= $this->ls[3] . ";";
            $linestring .= addslashes(($this->ls[4]));
        }
        $linestring .= "!__!\n"; //Zeilenende ist nicht eindeutig

        if ($this->outfile) {
            fwrite($this->of, $linestring);
        } else if ($this->output_action) {
            $this->csv_buffer .= $linestring;
        }
    }

    /**
    * Interne Funktion. Schreibt in Datei falls vorhanden oder in String
    * Falls ein Einfgestring bergeben wurde, wird dieser eingefgt,
    * sonst wird die XML-Ausgabe von get_xml generiert.
    *
    * @param string $insert optional   String der eingefgt (zwischengeschaltet) wird
    * @access private
    */
    function write_xml($insert = "")
    {
        if (! $this->highlight) {
            $xmlstring = ($insert) ? $insert : $this->get_xml();
        } else {
            $xmlstring = ($insert) ? htmlentities($insert) : $this->get_hi_xml();
        }
        if ($this->outfile) {
            fwrite($this->of, $xmlstring);
        } else if ($this->output_action) {
            $this->xml_buffer .= $xmlstring;
        }
    }
    // ===========neu ==============
    function copy_xml()
    {
        $xmlstring = $this->get_xml();
        foreach($this->copy_flag as $k => $v) {
            $this->copy_flag[$k] .= $xmlstring;
        }
    }

    /**
    * Baut die XML-Ausgabe wieder zusammen.
    *
    * @access private
    */
    function get_xml()
    {
        if ($this->ls[0] == "S") {
            $el = strrchr($this->ls[2], '/');
            $element = (!$el) ? $this->ls[2] : substr($el, 1);
            return "<" . $element;
        } else if ($this->ls[0] == "E") {
            $el = strrchr($this->ls[2], '/');
            $element = (!$el) ? $this->ls[2] : substr($el, 1);
            return "</" . $element . ">";
        } else if ($this->ls[0] == "D") {
            // return $this->ls[4];
            return str_replace($this->tmp_sign, '', $this->ls[4]);
        } else if ($this->ls[0] == "A") {
            // wichtig: Single-Line-Modus falls Attribut getrennt wird.
            return ' ' . $this->ls[3] . '="' . $this->ls[4] . '"';
        } else if ($this->ls[0] == "EOS") {
            return '>'; //Tag schlieen
        }
    }

    /**
    * Verantwortlich fr die Ausgabe von XML als HTML mit Syntaxhighlighting mit span-Elementen und
    * class-Attributen. Diese Funktion wird aktiv wenn als Rckgabeformat nicht 'xml' sondern 'hixml'
    * angegeben wird. - Todo: eventuell Umbau fr <pre>-Ausgabe
    *
    * @access private
    */
    function get_hi_xml()
    {
        $sp = ""; //Space fr HTML
        if ($this->ls[0] == "S") {
            $el = strrchr($this->ls[2], '/');
            $element = (!$el) ? $this->ls[2] : substr($el, 1);
            $newline = "";
            if ($this->breakset == false) $newline = "<br />\r\n";
            $this->breakset = true;
            $this->datafound = false; //Merker zurcksetzen
            return $newline . $this->sethtmlspace() . "<span class='klammer'>&lt;</span><span class='elname'>" . $element . "</span>";
        } else if ($this->ls[0] == "E") {
            $el = strrchr($this->ls[2], '/');
            $element = (!$el) ? $this->ls[2] : substr($el, 1);
            if ($this->breakset == true) {
                $sp = $this->sethtmlspace();
            }
            $this->breakset = true;
            $this->datafound = false;
            return $sp . "<span class='klammer'>&lt;/</span><span class='elname'>" . $element . "</span><span class='klammer'>&gt;</span><br />\r\n";
        } else if ($this->ls[0] == "D") {
            // Elementinahlt in Variable packen und prfen, ob sie leer ist.
            // wenn ja, Flag setzen.
            $br = "";
            // Falls XML eingefgt wird
            if ($tmpdata = htmlentities(trim(str_replace($this->tmp_sign, '', $this->ls[4])))) {
                $this->datafound = true;
                if ($this->breakset == true) {
                    $sp = $this->sethtmlspace();
                    $br = "<br />\r\n";
                }
            } else {
                $this->datafound = false;
            }

            return $sp . $tmpdata . $br; //Space und Daten zurckgeben
        } else if ($this->ls[0] == "A") {
            // wichtig: Single-Line-Modus falls Attribut getrennt wird.
            return " <span class='attname'>" . $this->ls[3] . "</span>=&quot;<span class='attr'>" . trim($this->ls[4]) . "</span>&quot;";
        } else if ($this->ls[0] == "EOS") {
            $this->datafound = false;
            $this->breakset = false;
            return "<span class='klammer'>&gt;</span>";
        }
    }
    /**
    * Rckt die Highlight-Ausgabe ein. wird nur von der Funktion get_high_xml benutzt.
    *
    * @see get_hi_xml
    * @access private
    */
    function sethtmlspace()
    {
        $space = "";
        for ($i = 1; $i < $this->level; $i++) {
            $space .= " &nbsp; - &nbsp;";
        }
        return $space;
    }

    /**
    * Liefert den Output als String, wenn diese Variante gewhlt wurde.
    * Es wird geprft, ob entweder der XML- oder der CSV-Buffer Inhalt hat.
    *
    * @access public
    */
    function get_output()
    {
        if ($this->xml_buffer) {
            return $this->xml_buffer;
        } else if ($this->csv_buffer) {
            return $this->csv_buffer;
        } else {
            return false;
        }
    }
}

?>